from dataclasses import dataclass, field
from abc import ABC
    
@dataclass(frozen=True)
class Component(ABC):
    nodes : tuple[str]
    type : str
    id : str

@dataclass(frozen=True)
class TwoPoleComponent(Component):
    nodes : tuple[str, str]

@dataclass(frozen=True)
class Resistor(TwoPoleComponent):
    R : float
    type : str = field(default='resistor', init=False)

@dataclass(frozen=True)
class Capacitor(TwoPoleComponent):
    C : float
    type : str = field(default='capacitor', init=False)

@dataclass(frozen=True)
class Inductance(TwoPoleComponent):
    L : float
    type : str = field(default='inductance', init=False)

@dataclass(frozen=True)
class Source(TwoPoleComponent):
    w : float = field(default=0.0)
    phi : float = field(default=0.0)

@dataclass(frozen=True)
class CurrentSource(Source):
    I : float = field(default=0.0)
    type : str = field(default='current_source', init=False)

@dataclass(frozen=True)
class VoltageSource(Source):
    V : float = field(default=0.0)
    type : str = field(default='voltage_source', init=False)

@dataclass(frozen=True)
class Ground(Component):
    id : str = field(default='gnd')
    type : str = field(default='ground', init=False)
    nodes : tuple[str]


from ..Utils import ScientificComplex

blue = '#02468F'
red = '#D20000'

def print_current(I: complex, precision: int = 3, polar=False) -> str:
    return str(ScientificComplex(I, 'A', precision=precision, polar=polar, use_exp_prefix=True, exp_prefixes={-6: 'u', -3: 'm', 3: 'k'}, compact=True))

def print_voltage(V: complex, precision: int = 3, polar=False) -> str:
    return str(ScientificComplex(V, 'V', precision=precision, polar=polar, use_exp_prefix=True, exp_prefixes={-6: 'u', -3: 'm', 3: 'k'}, compact=True))


import numpy as np
from ..Network import elements
from ..Network.network import Branch
from typing import Callable, Type, TypeVar
from . import components as cmp

CircuitComponent = TypeVar("CircuitComponent", bound=cmp.Component)
CircuitComponentTranslator = Callable[[CircuitComponent, float, float], Branch]

def resistor(resistor: cmp.Resistor, *_) -> Branch:
    return Branch(resistor.nodes[0], resistor.nodes[1], elements.resistor(resistor.id, resistor.R))

def capacitor(capacitor: cmp.Capacitor, w: float = 0, *_) -> Branch:
    return Branch(
        capacitor.nodes[0],
        capacitor.nodes[1],
        elements.admittance(capacitor.id, elements.admittance_value(B=w*capacitor.C)))

def inductance(inductance: cmp.Inductance, w: float = 0, *_) -> Branch:
    return Branch(
        inductance.nodes[0],
        inductance.nodes[1],
        elements.impedance(inductance.id, elements.impedance_value(X=w*inductance.L))
    )

def current_source(current_source: cmp.CurrentSource, w: float = 0, w_resolution: float = 1e-3) -> Branch:
    element = elements.current_source(current_source.id, elements.complex_value(current_source.I, 0))
    if np.abs(w-current_source.w) > w_resolution:
        element = elements.open_circuit(current_source.id)
    return Branch(
        current_source.nodes[0],
        current_source.nodes[1],
        element
    )

def voltage_source(voltage_source: cmp.VoltageSource, w: float = 0, w_resolution: float = 1e-3) -> Branch:
    element = elements.voltage_source(voltage_source.id, elements.complex_value(voltage_source.V, 0))
    if np.abs(w-voltage_source.w) > w_resolution:
        element = elements.short_cicruit(voltage_source.id)
    return Branch(
        voltage_source.nodes[0],
        voltage_source.nodes[1],
        element
    )

transformers : dict[Type[cmp.Component], CircuitComponentTranslator] = {
    cmp.Resistor : resistor,
    cmp.Capacitor : capacitor,
    cmp.Inductance : inductance,
    cmp.CurrentSource : current_source,
    cmp.VoltageSource : voltage_source
}


from .Network.network import Network
from .Network.transformers import switch_ground_node, passive_network
from .Network import labelmapper as map
from .NodalAnalysis import NodalAnalysisSolution, create_node_matrix_from_network
from numpy.linalg import inv as inverse_matrix
        
def calculate_total_impedeance(network: Network, node1: str, node2: str, node_index_mapper: map.NodeIndexMapper = map.default) -> complex:
    if node1 == node2:
        return 0
    if network.is_zero_node(node1):
        node1, node2 = node2, node1
    network = switch_ground_node(network=network, new_ground=node2)
    network = passive_network(network)
    Y = create_node_matrix_from_network(network, node_index_mapper=node_index_mapper)
    Z = inverse_matrix(Y)
    i1 = node_index_mapper(network)[node1]
    return Z[i1][i1]

def calculate_open_circuit_voltage(network: Network, node1: str, node2: str) -> complex:
    if node1 == node2:
        return 0
    solution = NodalAnalysisSolution(network)
    open_circuit_branch = network.branches_between(node1, node2)[0]
    if open_circuit_branch.node1 == node1:
        return solution.get_voltage(open_circuit_branch.id)
    else:
        return -solution.get_voltage(open_circuit_branch.id)

class TheveninEquivalentSource:
    def __init__(self, network: Network, node1: str, node2: str) -> None:
        self.U = calculate_open_circuit_voltage(network, node1, node2)
        self.Z = calculate_total_impedeance(network, node1, node2)

class NortenEquivalentSource:
    def __init__(self, network: Network, node1: str, node2: str) -> None:
        thevenin = TheveninEquivalentSource(network, node1, node2)
        self.I = thevenin.U/thevenin.Z
        self.Y = 1/thevenin.Z
        


from . import Elements as elm
from ..Circuit import components
from .SchemdrawTranslatorTypes import ElementTranslatorMap
import numpy as np


def resistor_translator(element: elm.Resistor, nodes: tuple[str, str]) -> components.Resistor:
    return components.Resistor(nodes=nodes, id=element.name, R=element.R)

def current_source_translator(element: elm.CurrentSource, nodes: tuple[str, str]) -> components.CurrentSource:
    return components.CurrentSource(nodes=nodes, id=element.name, I=element.I)

def voltage_source_translator(element: elm.VoltageSource, nodes: tuple[str, str]) -> components.VoltageSource:
    return components.VoltageSource(nodes=nodes, id=element.name, V=element.V)

def capacitor_translator(element: elm.Capacitor, nodes: tuple[str, str]) -> components.Capacitor:
    return components.Capacitor(nodes=nodes, id=element.name, C=element.C)

def inductance_translator(element: elm.Inductance, nodes: tuple[str, str]) -> components.Inductance:
    return components.Inductance(nodes=nodes, id=element.name, L=element.L)

def ground_translator(element: elm.Ground, nodes: tuple[str]) -> components.Ground:
    return components.Ground(nodes=nodes, id=element.name)

def ac_voltage_source_translator(element: elm.ACVoltageSource, nodes: tuple[str, str]) -> components.VoltageSource:
    return components.VoltageSource(
        nodes=nodes,
        id=element.name,
        V=element.V,
        w=element.w,
        phi=element.phi*np.pi/180 if element.deg else element.phi
    )

circuit_translator_map : ElementTranslatorMap = {
    elm.Resistor : resistor_translator,
    elm.CurrentSource : current_source_translator,
    elm.VoltageSource : voltage_source_translator,
    elm.Capacitor : capacitor_translator,
    elm.Inductance : inductance_translator,
    elm.Ground : ground_translator,
    elm.ACVoltageSource : ac_voltage_source_translator,
}



from .network import Network
from typing import Callable

NodeIndexMapper = Callable[[Network], dict[str, int]]

def alphabetic_mapper(network: Network) -> dict[str, int]:
    node_labels_without_zero = [label for label in sorted(network.node_labels) if label != network.zero_node_label] 
    return {k: v for v, k in enumerate(node_labels_without_zero)}

default = alphabetic_mapper


import schemdraw
import schemdraw.elements
from .Display import red, blue, print_voltage, print_current
from typing import Any
from ..Utils import ScientificFloat, ScientificComplex

def segments_of(element: schemdraw.elements.Element) -> list[schemdraw.segments.SegmentType]:
    return element.segments

class Schematic(schemdraw.Drawing):
    def __init__(self, unit=7, **kwargs):
        super().__init__(unit=unit, **kwargs)

    def save_copy(self, fname: str, **kwargs) -> None:
        import copy
        cpy = copy.deepcopy(self)
        cpy.save(fname, **kwargs)

class VoltageSource(schemdraw.elements.sources.Source):
    def __init__(self, V: complex, name: str, *args, reverse=False, precision=3, **kwargs):
        super().__init__(*args, reverse=reverse, **kwargs)
        if reverse:
            self._V = -V
        else:
            self._V = V
        self._name = name
        self.anchors['V_label'] = (0.5, 1.1)
        self.label(f'{self._name}={print_voltage(V, precision=precision)}', rotate=True, color=blue, loc='V_label', halign='center', valign='center')
        self.segments.append(schemdraw.segments.Segment([(0, 0), (1, 0)]))

        a, b = (1.5, 0.7), (-0.5, 0.7)
        self.segments.append(schemdraw.Segment((a, b), arrow='->', arrowwidth=.3, arrowlength=.4, color=blue))

    @property
    def name(self) -> str:
        return self._name

    @property
    def V(self) -> complex:
        return self._V

    def values(self) -> dict[str, complex]:
        return {'U' : self.V}

class ACVoltageSource(VoltageSource):
    def __init__(self, V: float, w: float, phi: float, name: str, *args, sin=False, deg=False, reverse=False, precision=3, label_offset: float = 0.2, **kwargs):
        super().__init__(V, name, *args, reverse=reverse, precision=precision, **kwargs)
        self._w = w
        self._phi = phi
        self._deg = deg
        self._sin = sin
        label = '$' + f'{self._V:4.2f}' + '\\mathrm{V}\\cdot\\cos(' + f'{self._w:4.2g}' + '\\cdot t + ' + f'{self._phi:4.2f}' + ')$'
        self.label(label, rotate=True, ofst=label_offset)

    @property
    def w(self) -> float:
        return self._w

    @property
    def phi(self) -> float:
        return self._phi

    @property
    def sin(self) -> bool:
        return self._sin

    @property
    def deg(self) -> bool:
        return self._deg

class CurrentSource(schemdraw.elements.sources.SourceI):
    def __init__(self, I: complex, name: str, *args, reverse=False, precision=3, **kwargs):
        super().__init__(*args, reverse=reverse, **kwargs)
        if reverse:
            self._I = -I
        else:
            self._I = I
        self._name = name
        self.segments = DrawCurrentSource()
        a, b = (1.2, 0.3), (1.8, 0.3)
        self.segments.append(schemdraw.Segment((a, b), arrow='->', arrowwidth=.3, arrowlength=.4, color=red))
        self.anchors['I_label'] = a
        self.label(f'{self._name}={print_current(self._I)}', loc='I_label', ofst=(0, 0.4), rotate=True, color=red)

    @property
    def name(self) -> str:
        return self._name

    @property
    def I(self) -> complex:
        return self._I

    def values(self) -> dict[str, complex]:
        return {'I' : self.I}

def DrawSource() -> list[schemdraw.segments.SegmentType]:
    return [
        schemdraw.segments.Segment([(0, 0), (0, 0), schemdraw.elements.elements.gap, (1, 0), (1, 0)]),
        schemdraw.segments.SegmentCircle((0.5, 0), 0.5,)
    ]

def DrawCurrentSource() -> list[schemdraw.segments.SegmentType]:
    return DrawSource() + [
        schemdraw.segments.Segment([(0.5, -0.5), (0.5, 0.5)])
    ]

def DrawRealCurrentSource() -> list[schemdraw.segments.SegmentType]:
    return DrawCurrentSource() + [
        schemdraw.segments.SegmentCircle((-0.5, 0), 0.07, fill='black'),
        schemdraw.segments.SegmentCircle((1.5, 0), 0.07, fill='black'),
        schemdraw.segments.Segment([(-0.5, 0), (-0.5, 1)]),
        schemdraw.segments.Segment([(1.5, 0), (1.5, 1)]),
        schemdraw.segments.Segment([(-0.5, 1), (0, 1)]),
        schemdraw.segments.Segment([(1, 1), (1.5, 1)]),
        schemdraw.segments.Segment([(0, 0.8), (1, 0.8), (1, 1.2), (0, 1.2), (0, 0.8)])
    ]

def DrawVoltageSource() -> list[schemdraw.segments.SegmentType]:
    return DrawSource() + [
        schemdraw.segments.Segment([(0, 0), (1, 0)])
    ]

class Impedance(schemdraw.elements.twoterm.ResistorIEC):
    def __init__(self, Z: complex, name: str, *args, show_name: bool = True, show_value: bool = True, precision: int = 3, **kwargs):
        super().__init__(*args, **kwargs)
        self._Z = Z
        self._name = name
        label = ''
        label += f'{self._name}' if show_name else ''
        label += '=' if  show_name and show_value else ''
        label += str(ScientificComplex(value=self.Z, unit='$\\Omega$', use_exp_prefix=True, precision=precision)) if show_value else ''
        self.anchors['Z_label'] = (0.5, 0.3)
        self.label(label, rotate=True, loc='Z_label', halign='center')

    def down(self) -> schemdraw.elements.Element:
        self.anchors['Z_label'] = (0.5, -0.9)
        return super().down()

    def left(self) -> schemdraw.elements.Element:
        self.anchors['Z_label'] = (0.5, -1)
        return super().left()

    @property
    def name(self) -> str:
        return self._name

    @property
    def Z(self) -> complex:
        return self._Z

    @property
    def Y(self) -> complex:
        return 1/self._Z

    def values(self) -> dict[str, complex]:
        return {'Z' : self._Z}

    def _place_label(self, *args, **kwargs):
        delta = self.end-self.start
        if abs(delta[1]) > abs(delta[0]): # portrait placing of resistor
            if delta[1] < 0:
                kwargs.update({'rotation': 90})

        super()._place_label(*args, **kwargs)

class Resistor(schemdraw.elements.twoterm.ResistorIEC):
    def __init__(self, R: float, name: str, *args, show_name: bool = True, show_value: bool = True, **kwargs):
        super().__init__(*args, **kwargs)
        self._R = R
        self._name = name
        label = ''
        label += f'{self._name}' if show_name else ''
        label += '=' if  show_name and show_value else ''
        label += str(ScientificFloat(value=self.R, unit='$\\Omega$', use_exp_prefix=True)) if show_value else ''
        self.anchors['R_label'] = (0.5, 0.3)
        self.label(label, rotate=True, loc='R_label', halign='center')

    def down(self) -> schemdraw.elements.Element:
        self.anchors['R_label'] = (0.5, -0.9)
        return super().down()

    def left(self) -> schemdraw.elements.Element:
        self.anchors['R_label'] = (0.5, -1)
        return super().left()

    @property
    def name(self) -> str:
        return self._name

    @property
    def R(self) -> float:
        return self._R

    @property
    def G(self) -> float:
        return 1/self._R

    def values(self) -> dict[str, float]:
        return {'R' : self._R}

    def _place_label(self, *args, **kwargs):
        delta = self.end-self.start
        if abs(delta[1]) > abs(delta[0]): # portrait placing of resistor
            if delta[1] < 0:
                kwargs.update({'rotation': 90})

        super()._place_label(*args, **kwargs)

class Capacitor(schemdraw.elements.twoterm.Capacitor):
    def __init__(self, C: float, name: str, *args, show_name: bool = True, show_value: bool = True, label_offset: float = 0.2, **kwargs):
        super().__init__(*args, **kwargs)
        self._C = C
        self._name = name
        label = ''
        label += f'{self._name}' if show_name else ''
        label += '=' if  show_name and show_value else ''
        label += str(ScientificFloat(value=self.C, unit='$\\mathrm{F}$', use_exp_prefix=True)) if show_value else ''
        self.label(label, rotate=True, ofst=label_offset)

    @property
    def name(self) -> str:
        return self._name

    @property
    def C(self) -> float:
        return self._C

    def values(self) -> dict[str, float]:
        return {'C' : self._C}

    def _place_label(self, *args, **kwargs):
        delta = self.end-self.start
        if abs(delta[1]) > abs(delta[0]): # portrait placing of resistor
            if delta[1] < 0:
                kwargs.update({'rotation': 90})

        super()._place_label(*args, **kwargs)

class Inductance(schemdraw.elements.twoterm.Inductor):
    def __init__(self, L: float, name: str, *args, show_name: bool = True, show_value: bool = True, label_offset: float = 0.2, **kwargs):
        super().__init__(*args, **kwargs)
        self._L = L
        self._name = name
        label = ''
        label += f'{self._name}' if show_name else ''
        label += '=' if  show_name and show_value else ''
        label += str(ScientificFloat(value=self.L, unit='$\\mathrm{H}$', use_exp_prefix=True)) if show_value else ''
        self.label(label, rotate=True, ofst=label_offset)

    @property
    def name(self) -> str:
        return self._name

    @property
    def L(self) -> float:
        return self._L

    def values(self) -> dict[str, float]:
        return {'L$' : self._L}

    def _place_label(self, *args, **kwargs):
        delta = self.end-self.start
        if abs(delta[1]) > abs(delta[0]): # portrait placing of resistor
            if delta[1] < 0:
                kwargs.update({'rotation': 90})

        super()._place_label(*args, **kwargs)

class RealCurrentSource(schemdraw.elements.Element2Term):
    def __init__(self, current_source: CurrentSource, resistor: Resistor, *args, zoom_resistor=0.7, **kwargs):
        super().__init__(*args, **kwargs)
        self.segments += segments_of(current_source)
        transform = schemdraw.transform.Transform(theta = 0, globalshift=((1-zoom_resistor)/2,-1), localshift=(0, 0), zoom=zoom_resistor)
        self.segments += [s.xform(transform) for s in segments_of(resistor)]
        left_line = schemdraw.Segment([(-1, 0), (-1, -1), ((1-zoom_resistor)/2, -1)])
        right_line = schemdraw.Segment([(2, 0), (2, -1), ((1+zoom_resistor)/2, -1)])
        self.segments += [left_line, right_line]
        self.anchors.update(current_source.anchors)
        self.anchors.update({k:(v[0]+0, v[1]-2.2) for k, v in resistor.anchors.items()})
        self._userlabels += current_source._userlabels
        self._userlabels += resistor._userlabels
        self._name = current_source.name
        self._I = current_source.I
        self._R = resistor.R

    @property
    def name(self) -> str:
        return self._name

    @property
    def I(self) -> complex:
        return self._I

    @property
    def R(self) -> float:
        return self._R

    @property
    def G(self) -> float:
        return 1/self._R

    def values(self) -> dict[str, complex]:
        return {'I' : self.I, 'R' : self.R}

class RealVoltageSource(schemdraw.elements.Element2Term):
    def __init__(self, voltage_source: VoltageSource, resistor: Resistor, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.segments.append(schemdraw.segments.Segment([(0, 0), (0, 0), schemdraw.elements.elements.gap, (4, 0), (4, 0)]))
        self.segments.extend(segments_of(voltage_source))
        self.segments.append(schemdraw.segments.Segment([(1, 0), (3, 0)]))
        transform = schemdraw.transform.Transform(theta = 0, globalshift=(3, 0))
        self.segments.extend([s.xform(transform) for s in segments_of(resistor)])
        self.anchors.update(voltage_source.anchors)
        self.anchors.update({k:(v[0]+3, v[1]) for k, v in resistor.anchors.items()})
        self._userlabels += voltage_source._userlabels
        self._userlabels += resistor._userlabels
        self._name = voltage_source.name
        self._V = voltage_source.V
        self._R = resistor.R

    @property
    def name(self) -> str:
        return self._name

    @property
    def V(self) -> complex:
        return self._V

    @property
    def R(self) -> float:
        return self._R

    @property
    def G(self) -> float:
        return 1/self._R

    def values(self) -> dict[str, float]:
        return {'I' : self.I, 'R' : self.R}

class Line(schemdraw.elements.lines.Line):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def name(self) -> str:
        return ''

class Node(schemdraw.elements.Element):
    def __init__(self, id: str = '', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.node_id = id
        self.params['theta'] = 0
        self.params['drop'] = (0, 0)
        self.anchors['start'] = (0, 0)
        self.anchors['center'] = (0, 0)
        self.anchors['end'] = (0, 0)
        self.anchors['NE'] = (0.5, 0.1)
        self.anchors['NW'] = (-0.5, 0.1)
        self.anchors['SE'] = (0.5, -0.3)
        self.anchors['SW'] = (-0.5, -0.3)

    @property
    def name(self) -> str:
        return f'Node {self.node_id}'

class LabelNode(Node):
    def __init__(self, id : str = '', id_loc : str | dict[str, Any] = '', *args, show=True, **kwargs):
        super().__init__(id, *args, **kwargs)
        locations = {
            'W': {'loc': 'left', 'halign': 'right', 'valign': 'center'},
            'N': {'loc': 'top', 'halign': 'center', 'valign': 'bottom'},
            'NE': {'loc': 'NE', 'halign': 'left', 'valign': 'bottom'},
            'NW': {'loc': 'NW', 'halign': 'right', 'valign': 'bottom'},
            'E': {'loc': 'right', 'halign': 'left', 'valign': 'center'},
            'S': {'loc': 'bottom', 'halign': 'center', 'valign': 'top'},
            'SW': {'loc': 'SW', 'halign': 'left', 'valign': 'top'},
            'SE': {'loc': 'SE', 'halign': 'right', 'valign': 'top'}
        }
        self.segments.append(schemdraw.SegmentCircle((0, 0), 0.12, fill='black'))
        self.id_loc = {}
        if isinstance(id_loc, str):
            self.id_loc.update(locations.get(id_loc, {}))
        else:
            self.id_loc.update(id_loc)
        if show:
            self.show()

    def show(self):
        self.segments.append(schemdraw.SegmentCircle((0, 0), 0.12, fill='black'))
        self.bbox = self.get_bbox(includetext=False)
        self.label(f'{self.node_id}', **self.id_loc)

    @property
    def name(self) -> str:
        return f'Node {self.node_id}'

class Ground(Node):
    def __init__(self, id: str = '0', *args, **kwargs):
        super().__init__(id, *args, **kwargs)
        gndgap = 0.12
        gnd_lead = 0.4
        resheight = schemdraw.elements.twoterm.resheight
        gap = schemdraw.elements.elements.gap
        self.segments.append(schemdraw.Segment(
            [(0, 0), (0, -gnd_lead), (-resheight, -gnd_lead),
             (resheight, -gnd_lead), gap, (-resheight*.7, -gndgap-gnd_lead),
             (resheight*.7, -gndgap-gnd_lead), gap,
             (-resheight*.2, -gndgap*2-gnd_lead),
             (resheight*.2, -gndgap*2-gnd_lead)]))

    @property
    def name(self) -> str:
        return 'Ground'

class VoltageLabel(schemdraw.elements.CurrentLabel):
    def __init__(self, at: schemdraw.elements.Element, label: str = '', label_loc: str = 'bottom', **kwargs):
        kwargs['color'] = kwargs.get('color', blue)
        super().__init__(**kwargs)
        if isinstance(at, RealVoltageSource):
            self.at(at.center)
        else:
            try:
                self.at(at.v_label)
                self.theta(at.transform.theta)
            except AttributeError:
                self.at(at)
        self.label(label, rotate=kwargs.get('rotate', True), loc=label_loc, ofst=(0, -0.1))

class CurrentLabel(schemdraw.elements.CurrentLabelInline):
    def __init__(self, at: schemdraw.elements.Element, label: str = '', **kwargs):
        kwargs['color'] = kwargs.get('color', red)
        totlen = at._userparams.get('l', at._userparams.get('unit', 3))
        kwargs['ofst'] = totlen/4-0.15+kwargs.get('ofst', 0)
        start = kwargs.get('start', True)
        reverse = kwargs.get('reverse', False)
        kwargs.update({'start' : start, 'reverse' : reverse})
        super().__init__(**kwargs)
        self.at(at)
        self.label(label)

    @property
    def name(self) -> str:
        return ''

v_label_args : dict[Any, dict[str, Any]]= {
    Resistor : {'ofst' : -0.6},
    Impedance : {'ofst' : -0.6},
    CurrentSource : {'ofst' : 1.5, 'label_loc': 'top'},
    RealCurrentSource : {'ofst' : 1.5, 'label_loc': 'top'}
}

i_label_args : dict[Any, dict[str, Any]]= {
    Resistor : {'ofst' : 1.4},
    Impedance : {'ofst' : 1.4},
    VoltageSource : {'ofst' : -2.8},
    RealVoltageSource: {'ofst' : -0.8}
}


from CircuitCalculator.Network.loaders import load_network_from_json
from CircuitCalculator.NodalAnalysis import nodal_analysis_solver
import numpy as np

def test_network_1_with_advanced_nodal_analysis() -> None:
    network = load_network_from_json('./examples/networks/json/example_network_1.json')
    solution = nodal_analysis_solver(network)
    np.testing.assert_almost_equal(solution.get_voltage('R1'), 7.69, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R2'), 15.38, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('I1'), -23.08, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R1'), 0.77, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R2'), 0.77, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('I1'), -0.23, decimal=2)

def test_network_2_with_advanced_nodal_analysis() -> None:
    network = load_network_from_json('./examples/networks/json/example_network_2.json')
    solution = nodal_analysis_solver(network)
    np.testing.assert_almost_equal(solution.get_voltage('R1'), 1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R2'), 0.40, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R3'), 0.60, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('U1'), 1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R1'), 0.10, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R2'), 0.02, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R3'), 0.02, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('U1'), 0.12, decimal=2)

def test_network_3_with_advanced_nodal_analysis() -> None:
    network = load_network_from_json('./examples/networks/json/example_network_3.json')
    solution = nodal_analysis_solver(network)
    np.testing.assert_almost_equal(solution.get_voltage('R1'), 0.56, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R2'), 0.44, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R3'), 1.04, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R4'), 1.39, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R5'), -2.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('U1'), 1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('U2'), 2.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R1'), 0.056, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R2'), 0.022, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R3'), 0.035, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R4'), 0.035, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R5'), -0.04, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('U1'), 0.056, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('U2'), 0.075, decimal=3)

def test_network_4_with_advanced_nodal_analysis() -> None:
    network = load_network_from_json('./examples/networks/json/example_network_4.json')
    solution = nodal_analysis_solver(network)
    np.testing.assert_almost_equal(solution.get_voltage('R1'), 0.52, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R2'), 0.48, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R3'), -2.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R4'), 1.10, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R5'), 1.38, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('U1'), 1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('U2'), 2.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R1'), 0.052, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R2'), 0.024, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R3'), -0.067, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R4'), 0.028, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R5'), 0.028, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('U1'), 0.052, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('U2'), 0.094, decimal=3)

def test_network_5_with_advanced_nodal_analysis() -> None:
    network = load_network_from_json('./examples/networks/json/example_network_5.json')
    solution = nodal_analysis_solver(network)
    np.testing.assert_almost_equal(solution.get_voltage('R1'), -1.56, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R2'), 2.14, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R3'), 1.44, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R4'), 0.30, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('U1'), 1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('U2'), 2.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('U3'), 3.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('I4'), -2.7, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R1'), -0.156, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R2'), 0.107, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R3'), 0.048, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R4'), 0.0075, decimal=4)
    np.testing.assert_almost_equal(solution.get_current('U1'), 0.156, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('U2'), 0.048, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('U3'), 0.0075, decimal=4)
    np.testing.assert_almost_equal(solution.get_current('I4'), 0.1, decimal=3)

def test_network_8_with_advanced_nodal_analysis() -> None:
    network = load_network_from_json('./examples/networks/json/example_network_8.json')
    solution = nodal_analysis_solver(network)
    np.testing.assert_almost_equal(solution.get_voltage('R1'), 1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R2'), 2.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R3'), -0.75, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R4'), -4.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R5'), 3.75, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('U1'), 1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('U2'), 2.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('I3'), -7.75, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R1'), 0.10, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R2'), 0.10, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R3'), -0.025, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R4'), -0.10, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R5'), 0.075, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('U1'), 0.075, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('U2'), 0.075, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('I3'), 0.10, decimal=3)

def test_network_9_with_advanced_nodal_analysis() -> None:
    network = load_network_from_json('./examples/networks/json/example_network_9.json')
    solution = nodal_analysis_solver(network)
    np.testing.assert_almost_equal(solution.get_voltage('R1'), 0, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R2'), 20, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R3'), -20, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R4'), 100, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('Iq'), 0, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('Uq'), 120, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R1'), 0, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R2'), 4, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R3'), -1, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R4'), 5, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('Iq'), 4, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('Uq'), 5, decimal=3)



from .components import Component
from .transformers import transformers
from ..Network.network import Network
import numpy as np

Circuit = list[Component]

def w(f: float) -> float:
    return 2*np.pi*f

def transform(circuit: Circuit, w: float = 0, w_resolution: float = 1e-3) -> Network:
    ground_nodes = [component for component in circuit if component.type == 'ground']
    return Network(
        [transformers[type(component)](component, w, w_resolution) for component in circuit if type(component) in transformers.keys()],
        zero_node_label=ground_nodes[0].nodes[0]
    )


import numpy as np
from .Elements import complex_pointer
from .Layout import Layout, color, grid_layout
from functools import partial
from CircuitCalculator.Network.solution import NetworkSolution

class PointerDiagram:
    def __init__(self, layout: Layout = grid_layout, arrow_base: float = 0.05, arrow_length: float = 0.05):
        self.pointer_drawers = []
        self._max_length = 0
        self.fig, self.ax = layout()
        self._arrow_base = arrow_base
        self._arrow_length = arrow_length
        self.ax.set_aspect('equal', 'box')
        self.ax.set_xlabel(r'$\mathrm{Re}\,\{U\}$')
        self.ax.set_ylabel(r'$\mathrm{Im}\,\{U\}$')

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        for draw_pointer in self.pointer_drawers:
            draw_pointer(height=self._arrow_base*self._max_length, width=self._arrow_length*self._max_length)
        x_min, x_max = self.ax.get_xlim()
        y_min, y_max = self.ax.get_ylim()
        self.ax.set_xlim(xmin=min(x_min, y_min), xmax=max(x_max, y_max))
        self.ax.set_ylim(ymin=min(x_min, y_min), ymax=max(x_max, y_max))
        self.ax.legend(
            handles=[line for line in self.ax.lines],
            ncols=len(self.ax.lines),
            loc='upper center',
            bbox_to_anchor=(0.5, 1.1),
            frameon=False
        )

    def add_pointer(self, z: complex, z0: complex = 0, **kwargs):
        self._max_length = max(self._max_length, np.abs(z))
        self.pointer_drawers.append(partial(complex_pointer, self.ax, z0, z0+z, **kwargs))

class VoltagePointerDiagram(PointerDiagram):
    def __init__(self, solution: NetworkSolution, resistor: float = 1.0, **kwargs):
        self._solution = solution
        self._resistor = resistor
        self._pointer_heads = {}
        super().__init__(**kwargs)

    def _set_reference(self, ref_id: str, z: complex, z0: complex) -> None:
        self._pointer_heads.update({ref_id: z+z0})

    def _get_reference(self, ref_id: str) -> complex:
        if ref_id == '':
            return 0
        return self._pointer_heads.get(ref_id, self._solution.get_voltage(ref_id))

    def add_voltage_pointer(self, id: str, origin: str='', color=color['blue']) -> None:
        z = self._solution.get_voltage(id)
        z0 = self._get_reference(origin)
        self._set_reference(id, z, z0)
        self.add_pointer(z=z, z0=z0, color=color, label=f'$V({id})$')

    def add_current_pointer(self, id: str, color=color['red'], resistor: float = 0) -> None:
        if resistor == 0:
            resistor = self._resistor
        z = self._solution.get_current(id)*resistor
        self.add_pointer(z=z, color=color, label=f'$I({id})\cdot{resistor}\Omega$')

class CurrentPointerDiagram(PointerDiagram):
    def __init__(self, solution: NetworkSolution, conductor: float = 1.0, **kwargs):
        self._solution = solution
        self._conductor = conductor
        self._pointer_heads = {}
        super().__init__(**kwargs)

    def _set_reference(self, ref_id: str, z: complex, z0: complex) -> None:
        self._pointer_heads.update({ref_id: z+z0})

    def _get_reference(self, ref_id: str) -> complex:
        if ref_id == '':
            return 0
        return self._pointer_heads.get(ref_id, self._solution.get_current(ref_id))

    def add_current_pointer(self, id: str, origin: str='', color=color['red']) -> None:
        z = self._solution.get_current(id)
        z0 = self._get_reference(origin)
        self._set_reference(id, z, z0)
        self.add_pointer(z=z, z0=z0, color=color, label=f'$I({id})$')

    def add_voltage_pointer(self, id: str, color=color['blue'], conductor: float = 0) -> None:
        if conductor == 0:
            conductor = self._conductor
        z = self._solution.get_voltage(id)*conductor
        self.add_pointer(z=z, color=color, label=f'$V({id})\cdot{conductor}\mathrm{{S}}$')


import numpy as np
from .Network.network import Network
from .Network.elements import is_ideal_current_source, is_ideal_voltage_source, is_current_source
from .Network.transformers import passive_network
from .Network.supernodes import SuperNodes
from .Network import labelmapper as map
from .Network.solution import NetworkSolution
import itertools

class DimensionError(Exception): pass

def admittance_connected_to(network: Network, node: str) -> complex:
    return sum(b.element.Y for b in network.branches_connected_to(node) if np.isfinite(b.element.Y))

def admittance_between(network: Network, node1: str, node2: str) -> complex:
    return sum([b.element.Y for b in network.branches_between(node1, node2) if np.isfinite(b.element.Y)])

def create_node_matrix_from_network(network: Network, node_index_mapper: map.NodeIndexMapper = map.default) -> np.ndarray:
    super_nodes = SuperNodes(network)
    passive_net = passive_network(network)
    node_mapping = node_index_mapper(network)
    A = np.zeros((network.number_of_nodes-1, network.number_of_nodes-1), dtype=complex)
    def passive_node_pair(node1: str, node2: str) -> bool:
        return not super_nodes.is_active(node1) and not super_nodes.is_active(node2)
    def both_active(node1: str, node2: str) -> bool:
        return super_nodes.is_active(node1) and super_nodes.is_active(node2)
    def passive_matrix_element(node1: str, node2: str) -> complex:
        if node1 == node2:
            return admittance_connected_to(passive_net, node1)
        else:
            return -admittance_between(passive_net, node1, node2)
    def non_active_j_label() -> complex:
        Aij = -admittance_between(network, i_label, j_label)
        if super_nodes.is_reference(j_label):
            Aij -= admittance_between(network, i_label, super_nodes.get_active_node(j_label))
        if super_nodes.belong_to_same(i_label, j_label):
            Aij += admittance_connected_to(network, i_label)
        return Aij
    def matrix_element(i_label: str, j_label: str) -> complex:
        if passive_node_pair(i_label, j_label):
            return passive_matrix_element(i_label, j_label)
        if i_label == j_label:
            return -super_nodes.sign(i_label)
        if both_active(i_label, j_label) and super_nodes.belong_to_same(j_label, i_label):
            return super_nodes.sign(j_label)
        if not super_nodes.is_active(j_label):
            return non_active_j_label()
        return 0+0j
    for (i_label, i), (j_label, j) in itertools.product(node_mapping.items(), repeat=2):
        A[i, j] = matrix_element(i_label, j_label)
    return A

def create_current_vector_from_network(network: Network, node_index_mapper: map.NodeIndexMapper = map.default) -> np.ndarray:
    super_nodes = SuperNodes(network)
    b = np.zeros(network.number_of_nodes-1, dtype=complex)
    node_mapping = node_index_mapper(network)
    active_node_labels = [l for l in node_mapping.keys() if super_nodes.is_active(l)]
    def current_sources(node: str) -> complex:
        current_sources = [b for b in network.branches_connected_to(node) if is_current_source(b.element)]
        return sum([-cs.element.I if cs.node1 == node else cs.element.I for cs in current_sources])
    def connected_active_nodes(active_node: str) -> list[str]:
        return [n for n in network.nodes_connected_to(active_node) if super_nodes.is_active(n)]
    def active_node_current(active_node: str) -> complex:
        I_connected_admittances = super_nodes.voltage_to_next_reference(active_node)*admittance_connected_to(network, active_node)
        reference_node = super_nodes.get_reference_node(active_node)
        if super_nodes.is_active(reference_node):
            I_parallel_admittance = super_nodes.voltage_to_next_reference(reference_node)*admittance_between(network, active_node, reference_node)
            return I_parallel_admittance-I_connected_admittances
        return sum([super_nodes.voltage_to_next_reference(cn)*admittance_between(network, active_node, cn) for cn in connected_active_nodes(active_node)])-I_connected_admittances
    def passive_node_current(passive_node: str) -> complex:
        I_parallel_admittances = sum([super_nodes.voltage_to_next_reference(an)*admittance_between(network, an, passive_node) for an in active_node_labels if super_nodes.get_reference_node(an) != passive_node])
        if super_nodes.is_reference(passive_node):
            active_node = super_nodes.get_active_node(passive_node)
            I_connected_admittances = super_nodes.voltage_to_next_reference(active_node)*admittance_connected_to(network, active_node)
            I_parallel_to_active_nodes = sum([super_nodes.voltage_to_next_reference(cn)*admittance_between(network, active_node, cn) for cn in connected_active_nodes(active_node)])
            I_parallel_to_passive_nodes = super_nodes.voltage_to_next_reference(active_node)*admittance_between(network, active_node, passive_node)
            return I_parallel_admittances+I_parallel_to_active_nodes+I_parallel_to_passive_nodes-I_connected_admittances
        return I_parallel_admittances
    def current_of_voltage_sources(node: str) -> complex:
        if super_nodes.is_active(node):
            return active_node_current(node)
        return passive_node_current(node)
    for i_label, i in node_mapping.items():
        b[i] = current_sources(i_label) + current_of_voltage_sources(i_label)
    return b

def calculate_node_voltages(Y : np.ndarray, I : np.ndarray) -> np.ndarray:
    if np.any(np.logical_not(np.isfinite(Y))):
        raise ValueError
    if np.any(np.logical_not(np.isfinite(I))):
        raise ValueError
    if Y.ndim != 2:
        raise DimensionError('dim error')
    if I.ndim != 1:
        raise DimensionError('dim error')
    m, n = Y.shape
    if n != m:
        raise DimensionError('dim error')
    return np.linalg.solve(Y, I)

class NodalAnalysisSolution:
    def __init__(self, network : Network, node_mapper: map.NodeIndexMapper = map.default) -> None:
        Y = create_node_matrix_from_network(network, node_index_mapper=node_mapper)
        I = create_current_vector_from_network(network, node_index_mapper=node_mapper)
        self._network = network
        self._solution_vector = calculate_node_voltages(Y, I)
        self._super_nodes = SuperNodes(network)
        self._node_mapping = node_mapper(network)

    def _calculate_potential_of_node(self, node: str) -> complex:
        V_active = 0+0j
        if self._super_nodes.is_active(node):
            V_active = self._super_nodes.voltage_to_next_reference(node)
            node = self._super_nodes.next_reference(node)
        if self._network.is_zero_node(node):
            return V_active
        return self._solution_vector[self._node_mapping[node]] + V_active

    def _select_active_node(self, branch_id: str) -> str:
        branch = self._network[branch_id]
        if self._super_nodes.is_active(branch.node1):
            return branch.node1
        return branch.node2
    
    def get_voltage(self, branch_id: str) -> complex:
        phi1 = self._calculate_potential_of_node(self._network[branch_id].node1)
        phi2 = self._calculate_potential_of_node(self._network[branch_id].node2)
        return phi1-phi2

    def get_current(self, branch_id: str) -> complex:
        branch_element = self._network[branch_id].element
        if is_ideal_current_source(branch_element):
            return branch_element.I
        if is_ideal_voltage_source(branch_element):
            return self._solution_vector[self._node_mapping[self._select_active_node(branch_id)]]
        return self.get_voltage(branch_id)/branch_element.Z

def nodal_analysis_solver(network) -> NetworkSolution:
    return NodalAnalysisSolution(network)



import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axes import Axes
from typing import Callable

color = {
    'black' : (0, 0, 0),
    'grey' : (.7, .7, .7),
    'lightgrey' : (.9, .9, .9),
    'darkgrey' : (.5, .5, .5),
    'darkdarkgrey' : (.2, .2, .2),
    'darkblue' : (.13, .22, .39),
    'lightblue' : (.4, .7, 1),
    'red' : (0.721, 0, 0),
    'darkred' : (0.8, 0, 0),
    'green' : (0, 0.45, 0.333),
    'blue' : (.18, .33, .59),
    'purple' : (.44, .19, .63),
    'white' : (1, 1, 1)
}

Layout = Callable[[], tuple[Figure, Axes]]

def default_layout(**kwargs) -> tuple[Figure, Axes]:
    return plt.subplots(ncols=1, nrows=1, **kwargs)

def grid_layout(grid: bool = True, **kwargs) -> tuple[Figure, Axes]:
    fig, ax = default_layout(**kwargs)
    ax.grid(visible=grid, zorder=-1)
    return fig, ax


from .network import Network
from typing import Callable, Protocol

class NetworkSolution(Protocol):
    def get_voltage(self, branch_id: str) -> complex: ...

    def get_current(self, branch_id: str) -> complex: ...

NetworkSolver = Callable[[Network], NetworkSolution]


from .network import Network, Branch
from .elements import NortenTheveninElement, is_ideal_current_source, is_ideal_voltage_source
from .supernodes import SuperNodes

def switch_ground_node(network: Network, new_ground: str) -> Network:
    return Network(network.branches, new_ground)

def remove_ideal_current_sources(network: Network, keep: list[NortenTheveninElement] = []) -> Network:
    return Network([b for b in network.branches if not is_ideal_current_source(b.element) or b.element in keep], zero_node_label=network.zero_node_label)

def remove_ideal_voltage_sources(network: Network, keep: list[NortenTheveninElement] = []) -> Network:
    branches = network.branches
    super_nodes = SuperNodes(network)
    voltage_sources = [b for b in network.branches if is_ideal_voltage_source(b.element)]
    voltage_sources = [vs for vs in voltage_sources if vs.element not in keep]
    short_circuit_nodes = [(vs.node1, vs.node2) if super_nodes.is_active(vs.node1) else (vs.node2, vs.node1) for vs in voltage_sources]
    for an, rn in short_circuit_nodes:
        branches = [Branch(rn, b.node2, b.element) if b.node1 == an else b for b in branches]
        branches = [Branch(b.node1, rn, b.element) if b.node2 == an else b for b in branches]
        branches = [b for b in branches if b.node1 != b.node2]
    return Network(branches, zero_node_label=network.zero_node_label)

def passive_network(network: Network, keep: list[NortenTheveninElement] = []) -> Network:
    return remove_ideal_voltage_sources(remove_ideal_current_sources(network, keep=keep), keep=keep)